module MeterpreterDetection_HTTPHeader;

export {
    # Type to record state of HTTP parsing
    type parsing_state: enum { IDLE, FOUND_LENGTH };

    type HTTP_Info: record {
        reply_count:  count           &default=0;
        status:       parsing_state   &default=IDLE;
        length:       count           &default=0;
    };

    # Tables to store suspected and confirmed meterpreter HTTP connections
    global http_suspected: table[MeterpreterDetection::conn_triplet] of count &create_expire=12hr;
    global http_verified: set[MeterpreterDetection::conn_triplet] &create_expire=12hr;

    # State tracking for all HTTP connections.
    global http_info: table[conn_id] of HTTP_Info &create_expire=60sec;

    global sampling_size = 3;

    # Type for HTTP check return
    type Header: record {
        guid:           string;
        encrypted:      bool;
        len:            count;
        request:        bool;
    };

    global empty_header = [$guid = "", $encrypted = F, $len = 0, $request = T];

    # Authorized values in meterpreter header
    global allowed_encrypt_flags = set(0,1);
    global allowed_mtpt_types = set(0,1,10,11);
}

const http_body_analysis_byte_depth = 64 &redef;

redef record HTTP::Info += {
	meterpreter_analyzed_traffic_length: count &optional;
};

# Functions for HTTP header filtering
function xor_string(a: string, b: string): count
    {
    return bytestring_to_count(a) ^ bytestring_to_count(b);
    }

function reservoir_sampling(cnt: count): bool
    {
    if (cnt <= sampling_size)
        return T;

    local test = rand(cnt-1) + 1;
    return test <= sampling_size;
    }

# Takes a length and raw payload, returns the meterpreter header information.
# If this is not a meterpreter header, return the empty header.

function check_header(len: count, payload: string): Header
    {
    if (len < 32)
        return empty_header;

    local xor_key = payload[:4];
    local encrypt_flag = xor_string(xor_key, payload[20:24]);
    local mtpt_type = xor_string(xor_key, payload[28:32]);

    if (encrypt_flag in allowed_encrypt_flags && mtpt_type in allowed_mtpt_types)
        {
        local hdr: Header;
        hdr =  [$guid = MeterpreterDetection::count_to_bytestring(xor_string(xor_key + xor_key, payload[4:12])) + MeterpreterDetection::count_to_bytestring(xor_string(xor_key + xor_key, payload[12:20])),
                $encrypted = (encrypt_flag == 1),
                $len = xor_string(xor_key,payload[24:28]),
                $request = (mtpt_type==0 || mtpt_type==10)];
        return hdr;
        }

    return empty_header;
    }

# Event handlers for HTTP
function http_header_helper(cid: conn_id, is_orig: bool, name: string, value: string)
    {
    # Check if this is a length header
    if ("CONTENT-LENGTH" != name)
        return;

    # Only consider replies - messages from the control server
    if (is_orig)
        return;

    # Filters
    if (MeterpreterDetection::id_to_triplet(cid) in http_verified)
        return;

    # Get length value
    local content_length = 0;
    local stripped_value = strip(value);

    if ( |stripped_value| > 0 && is_num(stripped_value) )
        content_length = to_count(stripped_value);

    # Check that the length is strictly positive
    if (content_length == 0)
        return;

    # Load state
    if (cid !in http_info)
        http_info[cid] = [$reply_count=0];

    local rec = http_info[cid];

    # If we are not in the idle state, drop this reply
    if (rec$status != IDLE)
        return;

    # Sampling - This is to avoid checking every single packet in long connections.
    # If we decide to check this packet, we switch to the FOUND_LENGTH state.
    rec$reply_count += 1;

    if (reservoir_sampling(rec$reply_count))
        {
        rec$length = content_length;
        rec$status = FOUND_LENGTH;
        }

    # Update table
    http_info[cid] = rec;
    }

# This event is meant to catch the content-length header to check against the value in the MTPT header
event http_header(c: connection, is_orig: bool, name: string, value: string)
    {
    http_header_helper(c$id, is_orig, name, value) ;
    }

# Check HTTP data
event http_entity_data_helper(f: fa_file, data: string)
    {
    for ( cid in f$conns ) 
        {

        local c: connection = f$conns[cid];

        if (!c?$http)
            next;

        if ( ! c$http?$meterpreter_analyzed_traffic_length )
               c$http$meterpreter_analyzed_traffic_length = 0;

        # If we are already above the captured size here, just return.
	if ( c$http$meterpreter_analyzed_traffic_length > http_body_analysis_byte_depth )
	     return;
        
        if (c$id !in http_info)
            next;

        # Filters
        if (MeterpreterDetection::id_to_triplet(c$id) in http_verified)
            next;

        local rec = http_info[c$id];

        # Check status
        if (rec$status != FOUND_LENGTH)
            next;
 
        # Change status
        rec$status = IDLE;

        # Check meterpreter header
        local length = |data|;
        local mtpt_hdr = check_header(length, data);
	local message: string;
        if (mtpt_hdr$len + 24 == rec$length && mtpt_hdr$len > 0)
            {
            # Raise Notice and add to suspected
            if (MeterpreterDetection::id_to_triplet(c$id) !in http_suspected)
                {
                http_suspected[MeterpreterDetection::id_to_triplet(c$id)] = 0;
                message=fmt("Meterpreter detected with reason=HEADER_DETECTED, protocol=HTTP, os=ANY");
                MeterpreterDetection::log_a_notice(c, message, MeterpreterDetection::Meterpreter_Detected, T);
                }

            # Check on the values of guid and encrypted flag, in case we are in an initialization phase. Same logi as TCP case
            local null_str: string = "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
            if ( (mtpt_hdr$encrypted && mtpt_hdr$guid != null_str) || 
                    (mtpt_hdr$guid != null_str && http_suspected[MeterpreterDetection::id_to_triplet(c$id)] > 3 ) ||
                    (http_suspected[MeterpreterDetection::id_to_triplet(c$id)] > 10) )
                {
                add http_verified[MeterpreterDetection::id_to_triplet(c$id)];
                message =fmt("Meterpreter header detected with protocol=HTTP in connection UID=%s, guid=%s, staged=T,"
                + "encrypted=%s", c$uid, string_to_ascii_hex(mtpt_hdr$guid), mtpt_hdr$encrypted);
                MeterpreterDetection::log_a_notice(c, message, MeterpreterDetection::Meterpreter_Header_Detected, T);
                }
            else
                {
                http_suspected[MeterpreterDetection::id_to_triplet(c$id)] += 1;
                rec$reply_count -= 1;
                rec$status = IDLE;
                }
            }

        # Update table
        http_info[c$id] = rec;
        c$http$meterpreter_analyzed_traffic_length += |data|;
	if ( c$http$meterpreter_analyzed_traffic_length > http_body_analysis_byte_depth )
	    {
	    Files::remove_analyzer(f, Files::ANALYZER_DATA_EVENT, [$stream_event=http_entity_data_helper]);
	    }
        }
    }

event file_over_new_connection(f: fa_file, c: connection, is_orig: bool)
	{
	if ( !is_orig && c?$http && c$http?$method )
		{
		Files::add_analyzer(f, Files::ANALYZER_DATA_EVENT, [$stream_event=http_entity_data_helper]);
		}
	}

